#ifndef MAKEDEPEND
#include <cstdio>
#endif

// A GameBoy CPU emulator
//
// Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)

#include "disasm.h"

inline void STZ();
inline void STC();
inline void STH();
inline void maskflags(const byte and, const byte or);
inline void maskflags(const byte and);

class bytereg
{
public:
	byte value;
	bytereg() : value(0)
	{
	}
	bytereg(byte v) : value(v)
	{
	}
	inline void affectflags(byte value)
	{
		if(!value)STZ();
		if((value>>4) < 9)STH();
	}
	byte operator++()                 { ++value; maskflags(0x0F);      affectflags(value); return value; }
	byte operator--()                 { --value; maskflags(0x0F,0x40); affectflags(value); return value; }
//	inline byte operator+(const bytereg &v)  { return value + v.value; }
//	inline byte operator-(const bytereg &v)  { return value - v.value; }
	byte operator+(byte v)
	{
		register byte tmp = value+v;
		maskflags(0x0F);
		if(tmp < value)STC();
		affectflags(tmp);
		return tmp;
	}
	byte operator-(byte v)
	{
		register byte tmp = value-v;
		maskflags(0x0F, 0x40);
		if(tmp > value)STC();
		affectflags(tmp);
		return tmp;
	}
	operator byte &()      { return value; }
	byte &operator() ()    { return value; }
}; 

static bytereg B,C,D,E, H,L, A,F, S,P;

class wordreg
{
	class bytereg tmp1, tmp2;
public:
	class bytereg &hi;
	class bytereg &lo;
	wordreg(bytereg &reg1, bytereg &reg2) : hi(reg1), lo(reg2)
	{
	}
	wordreg(word v) :tmp1(v>>8), tmp2((byte)v), hi(tmp1), lo(tmp2)
	{
	}
	wordreg &operator= (const wordreg &w)
	{
		set(w.value());
		return *this;
	}
	inline word value() const { return (hi << 8) | lo; }
	inline void set(word w) { hi = w>>8; lo = w; }
	word operator++()         { register word w=value()+1; set(w); return w; }
	word operator--()         { register word w=value()-1; set(w); return w; }
	word operator++(int)      { register word v=value(),w=v+1; set(w); return v; }
	word operator--(int)      { register word v=value(),w=v-1; set(w); return v; }
	word operator+(const wordreg &w)
	{
		/* ADD HL, rr */
		
		word v = value() + w.value();
		
		maskflags(0x8F);

		if(v < value()) STC();
			
		// todo: check bcd adjustment (half-carry-flag 0x20)
		
		return v;
	}
};

static wordreg BC(B,C), DE(D,E), HL(H,L), SP(S,P), AF(A,F);

inline bool ZF()
{
	return F & 0x80;
}
inline bool CF()
{
	return F & 0x10;
}

inline void CMC()
{
	F = (F & 0x9F) ^ 0x10;
}
inline void STZ()
{
	F |= 0x80;
}
inline void STC()
{
	F |= 0x10;
}
inline void STH()
{
	F |= 0x20;
}
inline void SCF()
{
	F = (F & 0x9F) | 0x10;
}

inline void maskflags(const byte and, const byte or)
{
	F.value = (F & and) | or;
}
inline void maskflags(const byte and)
{
	F.value &= and;
}
inline byte ADC(const bytereg &a, const bytereg &b)
{
	register byte res = a.value + b.value + CF();
	return res;
}
inline byte SBB(const bytereg &a, const bytereg &b)
{
	register byte res = a.value - b.value - CF();
	return res;
}
inline byte SHL(const bytereg &a)
{
	register byte res = a.value << 1;
	return res;
}
inline byte SHR(const bytereg &a)
{
	register byte res = a.value >> 1;
	return res;
}
inline byte SAR(const bytereg &a)
{
	register signed char c = a.value;
	c >>= 1;
	return c;
}
inline byte AND(const bytereg &a, const bytereg &b)
{
	register byte res = a.value & b.value;
	maskflags(0x0F, 0x20);
	if(!res)STZ();
	return res;
}
inline byte XOR(const bytereg &a, const bytereg &b)
{
	register byte res = a.value ^ b.value;
	maskflags(0x0F);
	if(!res)STZ();
	return res;
}
inline byte OR(const bytereg &a, const bytereg &b)
{
	register byte res = a.value | b.value;
	maskflags(0x0F);
	if(!res)STZ();
	return res;
}
inline void TEST(const bytereg &a, byte bitno)
{
	maskflags(0x10, 0x20);
	if(!(a.value & (1<<bitno)))STZ();	
}
inline byte BTS(const bytereg &a, byte bitno)
{
	return a.value | 1<<bitno;
}
inline byte BTC(const bytereg &a, byte bitno)
{
	return a.value & ~(1<<bitno);
}
inline byte DAA(const bytereg &a)
{
	register byte res = a.value;
	//0x14 + 0x27 = 0x3B
	//            ->0x41
	//0x14 - 0x27 = 0xED
	//            ->0xF3
	if(F() & 0x20)
	{
		maskflags(~0x20);
		res += 6;
	}
	return res;
}
inline byte NOT(const bytereg &a)
{
	register byte res = a.value ^ 255;
	maskflags(0x9F, 0x60);
	return res;
}
inline byte RCL(const bytereg &a)
{
	register byte res = a.value << 1;
	if(CF())res |= 1;
	maskflags(0x0F);
	if(a.value & 128)
	{
		STC();
		res |= 1;
	}
	if(!res)STZ();
	return res;		
}
inline byte RCR(const bytereg &a)
{
	register byte res = a.value >> 1;
	if(CF())res |= 0x80;
	maskflags(0x0F);
	if(a.value & 1)
	{
		STC();
		res |= 128;
	}
	if(!res)STZ();
	return res;
}
inline byte ROL(const bytereg &a)
{
	register byte res = a.value << 1;
 	maskflags(0x0F);
	if(a.value & 128)
	{
		STC();
		res |= 1;
	}
	if(!res)STZ();
	return res;		
}
inline byte ROR(const bytereg &a)
{
	register byte res = a.value >> 1;
	maskflags(0x0F);
	if(a.value & 1)
	{
		STC();
		res |= 128;
	}
	if(!res)STZ();
	return res;
}
inline byte SWAP(const bytereg &a)
{
	register byte res = a.value;
	res = ((res&15) << 4) | (res>>4);
 	maskflags(0x0F);
 	if(!res)STZ();
	return res;
}
inline void POKEW(const wordreg &w, const wordreg &v)
{
	word k = w.value();
	POKE(k,   v.lo);
	POKE(k+1, v.hi);
}
inline void JUMP(const wordreg &w)
{
	PC = w.value();
}
inline void RET()
{
	byte lo = PEEK(SP++);
	PC = (PEEK(SP++)<<8) | lo;
}
inline void IRET()
{
	RET();
	RETI();
}
inline void PUSH(const wordreg &w)
{
	POKE(--SP, w.hi);
	POKE(--SP, w.lo);
}
inline void POP(wordreg &w)
{
	w.lo = PEEK(SP++);
	w.hi = PEEK(SP++);
}
inline byte PEEK(const wordreg &w)
{	
	return PEEK(w.value());
}
/*
inline void POKE(const wordreg &w, const bytereg &b)
{	
	POKE(w.value(), b.value);
}
*/
inline void POKE(const wordreg &w, const byte b)
{	
	POKE(w.value(), b);
}
inline void CALL(const wordreg &w)
{
	PUSH(PC);
	PC = w.value();
}

unsigned emulate()
{
	register byte b0 = PEEK(PC++);
	register byte b1 = PEEK(PC);
	register byte b2 = PEEK(PC+1);

   	register unsigned short w = b1 | ((unsigned)b2 << 8);
   	
   	switch(b0)
   	{
   		case 0x00: break;
   		case 0x10: ++PC; STOP(b1); break;
   	
   		case 0x08: POKEW(w, SP); break;
   	
   		case 0x18: ++PC; JUMP(PC+(signed char)b1); break; // JMP disp
   		case 0xC3: PC+=2; JUMP(w); break; // JMP abs
   		case 0xE9: JUMP(HL); break;
   		case 0xC2: PC+=2; if(!ZF())JUMP(w); break;
   		case 0xD2: PC+=2; if(!CF())JUMP(w); break;
   		case 0xCA: PC+=2; if(ZF())JUMP(w); break;
   		case 0xDA: PC+=2; if(CF())JUMP(w); break;
   	
   		case 0x20: ++PC; if(!ZF())JUMP(PC+(signed char)b1); break; // JNZ disp
   		case 0x30: ++PC; if(!CF())JUMP(PC+(signed char)b1); break; // JNC disp
   		case 0x28: ++PC; if(ZF())JUMP(PC+(signed char)b1); break; // JZ disp
   		case 0x38: ++PC; if(CF())JUMP(PC+(signed char)b1); break; // JC disp
   	
   		case 0x01: B=C; break;       case 0x11: D=E; break;
   		case 0x21: H=L; break;       case 0x31: S=P; break;

   		case 0x03: ++BC; break;      case 0x13: ++DE; break;
   		case 0x0B: --BC; break;      case 0x1B: --DE; break;
   		case 0x23: ++HL; break;      case 0x33: ++SP; break;
   		case 0x2B: --HL; break;      case 0x3B: --SP; break;
   	
   		case 0x02: POKE(BC, A); break;      case 0x0A: A = PEEK(BC); break;
   		case 0x12: POKE(DE, A); break;      case 0x1A: A = PEEK(DE); break;
   		case 0x22: POKE(HL++, A); break;    case 0x2A: A = PEEK(HL++); break;
   		case 0x32: POKE(HL--, A); break;    case 0x3A: A = PEEK(HL--); break;
   	
   		case 0x04: ++B; break;                   case 0x05: --B; break;
   		case 0x0C: ++C; break;                   case 0x0D: --C; break;
   		case 0x14: ++D; break;                   case 0x15: --D; break;
   		case 0x1C: ++E; break;                   case 0x1D: --E; break;
   		case 0x24: ++H; break;                   case 0x25: --H; break;
   		case 0x2C: ++L; break;                   case 0x2D: --L; break;
   		case 0x3C: ++A; break;                   case 0x3D: --A; break;
   		case 0x34: { bytereg tmp(PEEK(HL)); POKE(HL, ++tmp); break; }
   		case 0x35: { bytereg tmp(PEEK(HL)); POKE(HL, --tmp); break; }
   	
   		case 0x06: ++PC; B=b1; break;
   		case 0x16: ++PC; D=b1; break;
   		case 0x26: ++PC; H=b1; break;
   		case 0x36: ++PC; POKE(HL, b1); break;
   	
   		case 0x07: A = ROL(A); break;   case 0x0F: A = ROR(A); break;
   		case 0x17: A = RCL(A); break;   case 0x1F: A = RCR(A); break;
   		case 0x27: A = DAA(A); break;   case 0x2F: A = NOT(A); break;
   		
   		case 0x37: SCF(); break; // Set CF
   		case 0x3F: CMC(); break; // Complement CF

   		case 0x09: HL = HL + BC; break;
   		case 0x19: HL = HL + DE; break;
   		case 0x29: HL = HL + HL; break;
   		case 0x39: HL = HL + SP; break;
   	
   		case 0x0E: ++PC; C=b1; break;
   		case 0x1E: ++PC; E=b1; break;
   		case 0x2E: ++PC; L=b1; break;
   		case 0x3E: ++PC; A=b1; break;
   	
   		case 0x40: B=B; break;         case 0x41: B=C; break;
 		case 0x42: B=D; break;         case 0x43: B=E; break;
   		case 0x44: B=H; break;         case 0x45: B=L; break;
   		case 0x46: B=PEEK(HL); break;  case 0x47: B=A; break;
   		case 0x48: C=B; break;         case 0x49: C=C; break;
   		case 0x4A: C=D; break;         case 0x4B: C=E; break;
   		case 0x4C: C=H; break;         case 0x4D: C=L; break;
   		case 0x4E: C=PEEK(HL); break;  case 0x4F: C=A; break;
   		case 0x50: D=B; break;         case 0x51: D=C; break;
   		case 0x52: D=D; break;         case 0x53: D=E; break;
   		case 0x54: D=H; break;         case 0x55: D=L; break;
   		case 0x56: D=PEEK(HL); break;  case 0x57: D=A; break;
   		case 0x58: E=B; break;         case 0x59: E=C; break;
   		case 0x5A: E=D; break;         case 0x5B: E=E; break;
   		case 0x5C: E=H; break;         case 0x5D: E=L; break;
   		case 0x5E: E=PEEK(HL); break;  case 0x5F: E=A; break;
   		case 0x60: H=B; break;         case 0x61: H=C; break;
   		case 0x62: H=D; break;         case 0x63: H=E; break;
   		case 0x64: H=H; break;         case 0x65: H=L; break;
   		case 0x66: H=PEEK(HL); break;  case 0x67: H=A; break;
   		case 0x68: L=B; break;         case 0x69: L=C; break;
   		case 0x6A: L=D; break;         case 0x6B: L=E; break;
   		case 0x6C: L=H; break;         case 0x6D: L=L; break;
   		case 0x6E: L=PEEK(HL); break;  case 0x6F: L=A; break;
   		case 0x78: A=B; break;         case 0x79: A=C; break;
   		case 0x7A: A=D; break;         case 0x7B: A=E; break;
   		case 0x7C: A=H; break;         case 0x7D: A=L; break;
   		case 0x7E: A=PEEK(HL); break;  case 0x7F: A=A; break;
   		case 0x70: POKE(HL, B); break;
   		case 0x71: POKE(HL, C); break;
   		case 0x72: POKE(HL, D); break;
   		case 0x73: POKE(HL, E); break;
   		case 0x74: POKE(HL, H); break;
   		case 0x75: POKE(HL, L); break;
   		case 0x76: HALT();         break; // wait for irq and save power
   		case 0x77: POKE(HL, A); break;
   	
   		case 0x80: A=A+B; break;           case 0x90: A=A-B; break;
   		case 0x81: A=A+C; break;           case 0x91: A=A-C; break;
   		case 0x82: A=A+D; break;           case 0x92: A=A-D; break;
   		case 0x83: A=A+E; break;           case 0x93: A=A-E; break;
   		case 0x84: A=A+H; break;           case 0x94: A=A-H; break;
   		case 0x85: A=A+L; break;           case 0x95: A=A-L; break;
   		case 0x86: A=A+PEEK(HL); break;    case 0x96: A=A-PEEK(HL); break;
   		case 0x87: A=A+A; break;           case 0x97: A=A-A; break;
   	
   		case 0x88: A=ADC(A,B); break;         case 0x98: A=SBB(A,B); break;
   		case 0x89: A=ADC(A,C); break;         case 0x99: A=SBB(A,C); break;
   		case 0x8A: A=ADC(A,D); break;         case 0x9A: A=SBB(A,D); break;
   		case 0x8B: A=ADC(A,E); break;         case 0x9B: A=SBB(A,E); break;
   		case 0x8C: A=ADC(A,H); break;         case 0x9C: A=SBB(A,H); break;
   		case 0x8D: A=ADC(A,L); break;         case 0x9D: A=SBB(A,L); break;
   		case 0x8E: A=ADC(A,PEEK(HL)); break;  case 0x9E: A=SBB(A,PEEK(HL));
   		case 0x8F: A=ADC(A,A); break;         case 0x9F: A=SBB(A,A); break;
   		
   		case 0xA0: A=AND(A,B); break;         case 0xA8: A=XOR(A,B); break;
   		case 0xA1: A=AND(A,C); break;         case 0xA9: A=XOR(A,C); break;
   		case 0xA2: A=AND(A,D); break;         case 0xAA: A=XOR(A,D); break;
   		case 0xA3: A=AND(A,E); break;         case 0xAB: A=XOR(A,E); break;
   		case 0xA4: A=AND(A,H); break;         case 0xAC: A=XOR(A,H); break;
   		case 0xA5: A=AND(A,L); break;         case 0xAD: A=XOR(A,L); break;
   		case 0xA6: A=AND(A,PEEK(HL)); break;  case 0xAE: A=XOR(A,PEEK(HL)); break;
   		case 0xA7: A=AND(A,A); break;         case 0xAF: A=XOR(A,A); break;
   	
   		case 0xB0: A=OR(A,B); break;          case 0xB8: A-B; break; // compare
   		case 0xB1: A=OR(A,C); break;		  case 0xB9: A-C; break;
   		case 0xB2: A=OR(A,D); break;		  case 0xBA: A-D; break;
   		case 0xB3: A=OR(A,E); break;		  case 0xBB: A-E; break;
   		case 0xB4: A=OR(A,H); break;		  case 0xBC: A-H; break;
   		case 0xB5: A=OR(A,L); break;		  case 0xBD: A-L; break;
   		case 0xB6: A=OR(A,PEEK(HL)); break;	  case 0xBE: A-PEEK(HL); break;
   		case 0xB7: A=OR(A,A); break;		  case 0xBF: A-A; break;
   	
   		case 0xC0: if(!ZF())RET(); break;
   		case 0xD0: if(!CF())RET(); break;
   		case 0xC8: if(ZF())RET(); break;
   		case 0xD8: if(CF())RET(); break;
   		
   		case 0xE0: ++PC; POKE(0xFF00|b1, A); break;
   		case 0xF0: ++PC; A = PEEK(0xFF00|b1); break;
   		case 0xE2: POKE(0xFF00|C(), A); break;
   		case 0xF2: A = PEEK(0xFF00|C()); break;
   	
   		case 0xC1: POP(BC); break;       case 0xC5: PUSH(BC); break;
   		case 0xD1: POP(DE); break;       case 0xD5: PUSH(DE); break;
   		case 0xE1: POP(HL); break;       case 0xE5: PUSH(HL); break;
   		case 0xF1: POP(AF); break;       case 0xF5: PUSH(AF); break;
   		
   		// D3, E3 = nothing
   		// E4, F4 = nothing
   		
   		case 0xF3: DISABLE(); break; // CLI
   		case 0xFB: ENABLE();  break; // STI
   	
   		case 0xC4: PC+=2; if(!ZF())CALL(w); break;
   		case 0xD4: PC+=2; if(!CF())CALL(w); break;
   		case 0xCC: PC+=2; if(ZF())CALL(w); break;
   		case 0xDC: PC+=2; if(CF())CALL(w); break;
   		case 0xCD: PC+=2; CALL(w); break;
   	
   		case 0xC6: ++PC; A=A+b1; break;
   		case 0xD6: ++PC; A=A-b1; break;
   		case 0xE6: ++PC; A=SBB(A, b1); break;
   		case 0xF6: ++PC; A=OR(A, b1); break;
   	
   		case 0xC7: CALL(0*8); break; // RST 00
   		case 0xCF: CALL(1*8); break; // RST 08
   		case 0xD7: CALL(2*8); break; // RST 10
   		case 0xDF: CALL(3*8); break; // RST 18
   		case 0xE7: CALL(4*8); break; // RST 20
   		case 0xEF: CALL(5*8); break; // RST 28
   		case 0xF7: CALL(6*8); break; // RST 30
   		case 0xFF: CALL(7*8); break; // RST 38

   		case 0xE8: SP = SP + (signed char)b1; break;
   		case 0xF8: HL = SP + (signed char)b1; break; // correct, both add sp
   	
   		case 0xC9: RET();  break;
   		case 0xD9: IRET(); break;
   		case 0xF9: SP=HL;  break;

   		case 0xEA: PC+=2; POKE(w, A); break;
   		case 0xFA: PC+=2; A = PEEK(w); break;
   	
   		// DB, EB = nothing
   		// EC, FC = nothing
   		// DD, ED, FD = nothing
   	
   		case 0xCE: ++PC; A = ADC(A, b1); break;
   		case 0xDE: ++PC; A = SBB(A, b1); break;
   		case 0xEE: ++PC; A = XOR(A, b1); break;
   		case 0xFE: ++PC; A - b1; break; // compare
   	
   		case 0xCB:
   		{
   			++PC;
   			if(b1 < 0x40)
   			{
   				switch(b1)
   				{
   					case 0x00: B = ROL(B); break;               case 0x08: B = ROR(B); break;
   					case 0x01: C = ROL(C); break;               case 0x09: C = ROR(C); break;
   					case 0x02: D = ROL(D); break;               case 0x0A: D = ROR(D); break;
					case 0x03: E = ROL(E); break;               case 0x0B: E = ROR(E); break;
					case 0x04: H = ROL(H); break;               case 0x0C: H = ROR(H); break;
					case 0x05: L = ROL(L); break;               case 0x0D: L = ROR(L); break;
					case 0x06: POKE(HL, ROL(PEEK(HL))); break;  case 0x0E: POKE(HL, ROR(PEEK(HL))); break;
					case 0x07: A = ROL(A); break;               case 0x0F: A = ROR(A); break;
					
					case 0x10: B = RCL(B); break;               case 0x18: B = RCR(B); break;
					case 0x11: C = RCL(C); break;               case 0x19: C = RCR(C); break;
					case 0x12: D = RCL(D); break;               case 0x1A: D = RCR(D); break;
					case 0x13: E = RCL(E); break;               case 0x1B: E = RCR(E); break;
					case 0x14: H = RCL(H); break;               case 0x1C: H = RCR(H); break;
					case 0x15: L = RCL(L); break;               case 0x1D: L = RCR(L); break;
					case 0x16: POKE(HL, RCL(PEEK(HL))); break;  case 0x1E: POKE(HL, RCR(PEEK(HL))); break;
					case 0x17: A = RCL(A); break;               case 0x1F: A = RCR(A); break;
					
					case 0x20: B = SHL(B); break;               case 0x28: B = SAR(A); break;
					case 0x21: C = SHL(C); break;               case 0x29: C = SAR(C); break;
					case 0x22: D = SHL(D); break;               case 0x2A: D = SAR(D); break;
					case 0x23: E = SHL(E); break;               case 0x2B: E = SAR(E); break;
					case 0x24: H = SHL(H); break;               case 0x2C: H = SAR(H); break;
					case 0x25: L = SHL(L); break;               case 0x2D: L = SAR(L); break;
					case 0x26: POKE(HL, SHL(PEEK(HL))); break;  case 0x2E: POKE(HL, SAR(PEEK(HL))); break;
					case 0x27: A = SHL(A); break;               case 0x2F: A = SAR(A); break;
					
					case 0x30: B = SWAP(B); break;              case 0x38: B = SHR(B); break;
					case 0x31: C = SWAP(C); break;              case 0x39: C = SHR(C); break;
					case 0x32: D = SWAP(D); break;              case 0x3A: D = SHR(D); break;
					case 0x33: E = SWAP(E); break;              case 0x3B: E = SHR(E); break;
					case 0x34: H = SWAP(H); break;              case 0x3C: H = SHR(H); break;
					case 0x35: L = SWAP(L); break;              case 0x3D: L = SHR(L); break;
					case 0x36: POKE(HL, SWAP(PEEK(HL))); break; case 0x3E: POKE(HL, SHR(PEEK(HL))); break;
					case 0x37: A = SWAP(A); break;              case 0x3F: A = SHR(A); break;
				}
			}
    		else
    		{
    			bytereg *tm = NULL;
    			register byte c = ((1+~b1) >> 3) & 7;
    			switch(b1&7)
    			{
    				case 0: tm = &B; break;
    				case 1: tm = &C; break;
    				case 2: tm = &D; break;
    				case 3: tm = &E; break;
    				case 4: tm = &H; break;
    				case 5: tm = &L; break;
    				case 7: tm = &A; break;
    			}
    			if(!tm)
    				switch(b1 >> 6) // 0x40, 0x80, 0xC0
    				{
        				case 1: TEST(PEEK(HL), c); break;
        				case 2: POKE(HL, BTS(PEEK(HL), c)); break;
        				case 3: POKE(HL, BTC(PEEK(HL), c));
        			}
    			else
    				switch(b1 >> 6)
    				{
        				case 1: TEST(*tm, c); break;
        				case 2: *tm = BTS(*tm, c); break;
        				case 3: *tm = BTC(*tm, c);
        			}
    		}
    		break;
    	}
    	
   		default:
			OPCODE(b0);
	}
	return PC;
}
